import logging
from typing import List

from tenacity import retry

from aci.resource._base import APIResource, retry_config
from aci.types.app_configurations import (
    AppConfiguration,
    AppConfigurationCreate,
    AppConfigurationsList,
)
from aci.types.enums import SecurityScheme

logger: logging.Logger = logging.getLogger(__name__)


class AppConfigurationsResource(APIResource):
    """Resource for managing app configurations."""

    @retry(**retry_config)  # type: ignore
    def list(
        self,
        app_names: list[str] | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> List[AppConfiguration]:
        """List app configurations.

        Args:
            app_names: Filter by app names.
            limit: Maximum number of results per response.
            offset: Pagination offset.

        Returns:
            List[AppConfiguration]: List of app configurations.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        validated_params = AppConfigurationsList(
            app_names=app_names,
            limit=limit,
            offset=offset,
        ).model_dump(exclude_none=True, mode="json")

        logger.info(f"Listing app configurations with params: {validated_params}")
        response = self._httpx_client.get(
            "app-configurations",
            params=validated_params,
        )

        data: List[dict] = self._handle_response(response)
        app_configurations = [AppConfiguration.model_validate(config) for config in data]
        return app_configurations

    @retry(**retry_config)  # type: ignore
    def get(self, app_name: str) -> AppConfiguration:
        """Get an app configuration by app name.

        Args:
            app_name: Name of the app to get configuration for.

        Returns:
            AppConfiguration: The app configuration.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        logger.info(f"Getting app configuration for app: {app_name}")
        response = self._httpx_client.get(f"app-configurations/{app_name}")
        data: dict = self._handle_response(response)
        app_configuration = AppConfiguration.model_validate(data)

        return app_configuration

    @retry(**retry_config)  # type: ignore
    def create(
        self,
        app_name: str,
        security_scheme: SecurityScheme,
    ) -> AppConfiguration:
        """Create an app configuration.

        Args:
            app_name: Unique name of the app to create configuration for.
            security_scheme: Security scheme to use for the app configuration.

        Returns:
            AppConfiguration: The created app configuration.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        # TODO: add support for security_scheme_overrides, all_functions_enabled, enabled_functions
        validated_params = AppConfigurationCreate(
            app_name=app_name,
            security_scheme=security_scheme,
            security_scheme_overrides=None,
            all_functions_enabled=True,
            enabled_functions=None,
        ).model_dump(exclude_none=True, mode="json")

        logger.info(f"Creating app configuration: {validated_params}")

        response = self._httpx_client.post(
            "app-configurations",
            json=validated_params,
        )
        data: dict = self._handle_response(response)

        return AppConfiguration.model_validate(data)

    @retry(**retry_config)  # type: ignore
    def delete(self, app_name: str) -> None:
        """Delete an app configuration.

        Args:
            app_name: Name of the app to delete configuration for.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        logger.info(f"Deleting app configuration for app: {app_name}")
        response = self._httpx_client.delete(f"app-configurations/{app_name}")
        self._handle_response(response)

    # TODO: update are not supported for now



import logging
from typing import List
from uuid import UUID

from httpx import Response
from tenacity import retry

from aci._constants import DEFAULT_AFTER_OAUTH2_FLOW_REDIRECT_URL
from aci.resource._base import APIResource, retry_config
from aci.types.enums import SecurityScheme
from aci.types.linked_accounts import (
    LinkedAccount,
    LinkedAccountAPIKeyCreate,
    LinkedAccountNoAuthCreate,
    LinkedAccountOAuth2Create,
    LinkedAccountsList,
    LinkedAccountUpdate,
)

logger: logging.Logger = logging.getLogger(__name__)


class LinkedAccountsResource(APIResource):
    """Resource for managing linked accounts."""

    @retry(**retry_config)  # type: ignore
    def list(
        self,
        app_name: str | None = None,
        linked_account_owner_id: str | None = None,
    ) -> List[LinkedAccount]:
        """List linked accounts.

        Args:
            app_name: Filter by app name.
            linked_account_owner_id: Filter by linked account owner ID.
            See https://www.aci.dev/docs/core-concepts/linked-account#what-is-linked-account-owner-id

        Returns:
            List[LinkedAccount]: List of linked accounts.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        params = LinkedAccountsList(
            app_name=app_name,
            linked_account_owner_id=linked_account_owner_id,
        ).model_dump(exclude_none=True, mode="json")

        logger.info(f"Listing linked accounts with params: {params}")
        response = self._httpx_client.get("linked-accounts", params=params)

        data: List[dict] = self._handle_response(response)
        linked_accounts = [LinkedAccount.model_validate(account) for account in data]

        return linked_accounts

    @retry(**retry_config)  # type: ignore
    def get(self, linked_account_id: UUID) -> LinkedAccount:
        """Get a linked account by its ID.

        Args:
            linked_account_id: ID of the linked account to get.
            Note: linked_account_id is different from the linked_account_owner_id.
            See https://www.aci.dev/docs/core-concepts/linked-account#what-is-linked-account-owner-id

        Returns:
            LinkedAccount: The linked account.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        logger.info(f"Getting linked account with linked_account_id: {linked_account_id}")
        response = self._httpx_client.get(f"linked-accounts/{linked_account_id}")
        data: dict = self._handle_response(response)
        linked_account = LinkedAccount.model_validate(data)

        return linked_account

    @retry(**retry_config)  # type: ignore
    def link(
        self,
        app_name: str,
        security_scheme: SecurityScheme,
        linked_account_owner_id: str,
        api_key: str | None = None,
    ) -> LinkedAccount | str:
        """Link an account with the specified authentication type.

        Args:
            app_name: Name of the app to link account for, e.g., "GMAIL"
            linked_account_owner_id: ID of the owner of the linked account, e.g., "johndoe"
            security_scheme: The security scheme to use for the linked account.
            api_key: API key for authentication (required when security_scheme is API_KEY).

        Returns:
            if security_scheme is API_KEY or NO_AUTH, returns the linked account.
            if security_scheme is OAUTH2, returns the OAuth2 authorization URL.

        Raises:
            ValueError: If required parameters for the specified security scheme are missing.
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        response: Response | None = None

        if security_scheme == SecurityScheme.API_KEY:
            if not api_key:
                raise ValueError("api_key parameter is required when security_scheme is API_KEY")
            validated_params = LinkedAccountAPIKeyCreate(
                app_name=app_name,
                linked_account_owner_id=linked_account_owner_id,
                api_key=api_key,
            ).model_dump(exclude_none=True, mode="json")

            logger.info(
                f"Creating linked account with API key for app: {app_name}, owner_id: {linked_account_owner_id}"
            )

            response = self._httpx_client.post("linked-accounts/api-key", json=validated_params)

            return LinkedAccount.model_validate(self._handle_response(response))

        elif security_scheme == SecurityScheme.NO_AUTH:
            validated_params = LinkedAccountNoAuthCreate(
                app_name=app_name,
                linked_account_owner_id=linked_account_owner_id,
            ).model_dump(exclude_none=True, mode="json")

            logger.info(
                f"Creating linked account with no auth for app: {app_name}, owner_id: {linked_account_owner_id}"
            )

            response = self._httpx_client.post("linked-accounts/no-auth", json=validated_params)

            return LinkedAccount.model_validate(self._handle_response(response))

        elif security_scheme == SecurityScheme.OAUTH2:
            validated_params = LinkedAccountOAuth2Create(
                app_name=app_name,
                linked_account_owner_id=linked_account_owner_id,
                # TODO: make this customizable for users?
                after_oauth2_link_redirect_url=DEFAULT_AFTER_OAUTH2_FLOW_REDIRECT_URL,
            ).model_dump(exclude_none=True, mode="json")

            logger.info(
                f"Creating linked account with OAuth2 for app: {app_name}, owner_id: {linked_account_owner_id}"
            )

            response = self._httpx_client.get("linked-accounts/oauth2", params=validated_params)
            response_data: dict[str, str] = self._handle_response(response)

            return response_data["url"]

    @retry(**retry_config)  # type: ignore
    def delete(self, linked_account_id: UUID) -> None:
        """Delete a linked account.

        Args:
            linked_account_id: ID of the linked account to delete.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        logger.info(f"Deleting linked account with ID: {linked_account_id}")
        response = self._httpx_client.delete(f"linked-accounts/{linked_account_id}")
        self._handle_response(response)

    @retry(**retry_config)  # type: ignore
    def disable(self, linked_account_id: UUID) -> LinkedAccount:
        """Disable a linked account.

        Args:
            linked_account_id: ID of the linked account to disable.

        Returns:
            LinkedAccount: The updated linked account.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        return self._update(linked_account_id, enabled=False)

    @retry(**retry_config)  # type: ignore
    def enable(self, linked_account_id: UUID) -> LinkedAccount:
        """Enable a linked account.

        Args:
            linked_account_id: ID of the linked account to enable.

        Returns:
            LinkedAccount: The updated linked account.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        return self._update(linked_account_id, enabled=True)

    def _update(self, linked_account_id: UUID, enabled: bool | None = None) -> LinkedAccount:
        """Update a linked account.

        Args:
            linked_account_id: ID of the linked account to update.
            enabled: whether to enable or disable the linked account.

        Returns:
            LinkedAccount: The updated linked account.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        validated_params = LinkedAccountUpdate(
            enabled=enabled,
        ).model_dump(exclude_none=True, mode="json")

        logger.info(f"Updating linked account with ID: {linked_account_id}")
        response = self._httpx_client.patch(
            f"linked-accounts/{linked_account_id}",
            json=validated_params,
        )
        data: dict = self._handle_response(response)

        return LinkedAccount.model_validate(data)



from __future__ import annotations

from typing import Any, Callable, Optional

from aci.types.functions import FunctionDefinitionFormat

from ._function_schema import DocstringStyle, function_schema

"""
This file is a modified version of the `tool.py` file from the `openai` library.
https://github.com/openai/openai-agents-python/blob/064e25b01b5c82c08aea66ff898ff27adbb013d8/src/agents/tool.py
It is used to convert a Python function to an LLM tool schema in the specified format.
"""


def to_json_schema(
    func: Callable[..., Any],
    format: FunctionDefinitionFormat,
    *,
    name_override: Optional[str] = None,
    description_override: Optional[str] = None,
    docstring_style: Optional[DocstringStyle] = None,
    use_docstring_info: bool = True,
) -> dict:
    """
    Convert a Python function to an LLM tool schema in the specified format.

    Args:
        func: The function to convert to a schema
        format: The schema format to generate, one of the following:
            - FunctionDefinitionFormat.OPENAI: for openai chat completions api
            - FunctionDefinitionFormat.OPENAI_RESPONSES: for openai responses api
            - FunctionDefinitionFormat.ANTHROPIC: for anthropic api
        name_override: Optional custom name for the function
        description_override: Optional custom description
        docstring_style: Optional docstring style for parsing
        use_docstring_info: Whether to use function docstring for descriptions

    Returns:
        A dictionary containing the tool schema in the requested format

    Examples:
        >>> def get_weather(location: str) -> str:
        ...     '''Get current temperature for a location.'''
        ...     return f"Weather information for {location}"
        ...
        >>> openai_schema = to_llm_schema(get_weather, FunctionDefinitionFormat.OPENAI)
        >>> openai_responses_schema = to_llm_schema(get_weather, FunctionDefinitionFormat.OPENAI_RESPONSES)
        >>> anthropic_schema = to_llm_schema(get_weather, FunctionDefinitionFormat.ANTHROPIC)
    """
    # Extract base function metadata
    base_schema = function_schema(
        func=func,
        name_override=name_override,
        description_override=description_override,
        docstring_style=docstring_style,
        use_docstring_info=use_docstring_info,
    )

    # Generate schema based on format
    if format == FunctionDefinitionFormat.OPENAI:
        return {
            "type": "function",
            "function": {
                "name": base_schema.name,
                "description": base_schema.description or "",
                "parameters": base_schema.params_json_schema,
            },
        }
    elif format == FunctionDefinitionFormat.OPENAI_RESPONSES:
        return {
            "type": "function",
            "name": base_schema.name,
            "description": base_schema.description or "",
            "parameters": base_schema.params_json_schema,
        }
    elif format == FunctionDefinitionFormat.ANTHROPIC:
        return {
            "name": base_schema.name,
            "description": base_schema.description or "",
            "input_schema": base_schema.params_json_schema,
        }
    else:
        # This should never happen due to type restrictions, but just in case
        raise ValueError(f"Unsupported schema format: {format}")



DEFAULT_MAX_RETRIES = 3
DEFAULT_RETRY_MULTIPLIER = 1
DEFAULT_RETRY_MIN_WAIT = 2
DEFAULT_RETRY_MAX_WAIT = 8
DEFAULT_SERVER_URL = "https://api.aci.dev/v1/"
DEFAULT_AFTER_OAUTH2_FLOW_REDIRECT_URL = "https://platform.aci.dev"



import uuid
from typing import Dict
from uuid import UUID

import httpx
import pytest
import respx

from aci import ACI
from aci._exceptions import (
    NotFoundError,
    ValidationError,
)
from aci.types.enums import SecurityScheme

from .utils import MOCK_BASE_URL

MOCK_LINKED_ACCOUNT_ID = str(uuid.uuid4())
MOCK_PROJECT_ID = str(uuid.uuid4())
MOCK_APP_NAME = "test-app"
MOCK_OWNER_ID = "test-owner-id"
MOCK_API_KEY = "test-api-key"


@respx.mock
@pytest.mark.parametrize(
    "list_params",
    [
        {},  # Base case: No optional parameters provided.
        {"app_name": MOCK_APP_NAME},  # Filter by app_name
        {"linked_account_owner_id": MOCK_OWNER_ID},  # Filter by linked_account_owner_id
        {"app_name": MOCK_APP_NAME, "linked_account_owner_id": MOCK_OWNER_ID},  # Filter by both
    ],
)
def test_list_linked_accounts_success(client: ACI, list_params: Dict) -> None:
    mock_response = [
        {
            "id": MOCK_LINKED_ACCOUNT_ID,
            "project_id": MOCK_PROJECT_ID,
            "app_name": MOCK_APP_NAME,
            "linked_account_owner_id": MOCK_OWNER_ID,
            "security_scheme": SecurityScheme.API_KEY,
            "enabled": True,
            "created_at": "2023-01-01T00:00:00Z",
            "updated_at": "2023-01-01T00:00:00Z",
        }
    ]

    route = respx.get(f"{MOCK_BASE_URL}linked-accounts").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    linked_accounts = client.linked_accounts.list(**list_params)
    assert len(linked_accounts) == 1
    assert linked_accounts[0].app_name == MOCK_APP_NAME
    assert linked_accounts[0].linked_account_owner_id == MOCK_OWNER_ID
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_get_linked_account_success(client: ACI) -> None:
    mock_response = {
        "id": MOCK_LINKED_ACCOUNT_ID,
        "project_id": MOCK_PROJECT_ID,
        "app_name": MOCK_APP_NAME,
        "linked_account_owner_id": MOCK_OWNER_ID,
        "security_scheme": SecurityScheme.API_KEY,
        "enabled": True,
        "created_at": "2023-01-01T00:00:00Z",
        "updated_at": "2023-01-01T00:00:00Z",
    }

    route = respx.get(f"{MOCK_BASE_URL}linked-accounts/{MOCK_LINKED_ACCOUNT_ID}").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    linked_account = client.linked_accounts.get(MOCK_LINKED_ACCOUNT_ID)
    assert linked_account.id == UUID(MOCK_LINKED_ACCOUNT_ID)
    assert linked_account.app_name == MOCK_APP_NAME
    assert linked_account.linked_account_owner_id == MOCK_OWNER_ID
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_get_linked_account_not_found(client: ACI) -> None:
    route = respx.get(f"{MOCK_BASE_URL}linked-accounts/{MOCK_LINKED_ACCOUNT_ID}").mock(
        return_value=httpx.Response(404, json={"message": "Linked account not found"})
    )

    with pytest.raises(NotFoundError) as exc_info:
        client.linked_accounts.get(MOCK_LINKED_ACCOUNT_ID)

    assert "Linked account not found" in str(exc_info.value)
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_create_with_api_key_success(client: ACI) -> None:
    mock_response = {
        "id": MOCK_LINKED_ACCOUNT_ID,
        "project_id": MOCK_PROJECT_ID,
        "app_name": MOCK_APP_NAME,
        "linked_account_owner_id": MOCK_OWNER_ID,
        "security_scheme": SecurityScheme.API_KEY,
        "enabled": True,
        "created_at": "2023-01-01T00:00:00Z",
        "updated_at": "2023-01-01T00:00:00Z",
    }

    route = respx.post(f"{MOCK_BASE_URL}linked-accounts/api-key").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    result = client.linked_accounts.link(
        app_name=MOCK_APP_NAME,
        linked_account_owner_id=MOCK_OWNER_ID,
        security_scheme=SecurityScheme.API_KEY,
        api_key=MOCK_API_KEY,
    )

    assert result.id == UUID(MOCK_LINKED_ACCOUNT_ID)
    assert result.app_name == MOCK_APP_NAME
    assert result.security_scheme == SecurityScheme.API_KEY
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_create_with_no_auth_success(client: ACI) -> None:
    mock_response = {
        "id": MOCK_LINKED_ACCOUNT_ID,
        "project_id": MOCK_PROJECT_ID,
        "app_name": MOCK_APP_NAME,
        "linked_account_owner_id": MOCK_OWNER_ID,
        "security_scheme": SecurityScheme.NO_AUTH,
        "enabled": True,
        "created_at": "2023-01-01T00:00:00Z",
        "updated_at": "2023-01-01T00:00:00Z",
    }

    route = respx.post(f"{MOCK_BASE_URL}linked-accounts/no-auth").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    result = client.linked_accounts.link(
        app_name=MOCK_APP_NAME,
        linked_account_owner_id=MOCK_OWNER_ID,
        security_scheme=SecurityScheme.NO_AUTH,
    )

    assert result.id == UUID(MOCK_LINKED_ACCOUNT_ID)
    assert result.app_name == MOCK_APP_NAME
    assert result.security_scheme == SecurityScheme.NO_AUTH
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_create_with_oauth2_success(client: ACI) -> None:
    mock_response = {"url": "https://example.com/oauth2/authorize"}

    route = respx.get(f"{MOCK_BASE_URL}linked-accounts/oauth2").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    result = client.linked_accounts.link(
        app_name=MOCK_APP_NAME,
        linked_account_owner_id=MOCK_OWNER_ID,
        security_scheme=SecurityScheme.OAUTH2,
    )

    assert result == "https://example.com/oauth2/authorize"
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_create_api_key_validation_error(client: ACI) -> None:
    with pytest.raises(ValueError) as exc_info:
        client.linked_accounts.link(
            app_name=MOCK_APP_NAME,
            linked_account_owner_id=MOCK_OWNER_ID,
            security_scheme=SecurityScheme.API_KEY,
            # missing api_key
        )

    assert "api_key parameter is required" in str(exc_info.value)


@respx.mock
def test_create_with_api_key_bad_request(client: ACI) -> None:
    route = respx.post(f"{MOCK_BASE_URL}linked-accounts/api-key").mock(
        return_value=httpx.Response(400, json={"message": "Bad request"})
    )

    with pytest.raises(ValidationError) as exc_info:
        client.linked_accounts.link(
            app_name=MOCK_APP_NAME,
            linked_account_owner_id=MOCK_OWNER_ID,
            security_scheme=SecurityScheme.API_KEY,
            api_key=MOCK_API_KEY,
        )

    assert "Bad request" in str(exc_info.value)
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_delete_linked_account_success(client: ACI) -> None:
    route = respx.delete(f"{MOCK_BASE_URL}linked-accounts/{MOCK_LINKED_ACCOUNT_ID}").mock(
        return_value=httpx.Response(204)
    )

    client.linked_accounts.delete(MOCK_LINKED_ACCOUNT_ID)
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_delete_linked_account_not_found(client: ACI) -> None:
    route = respx.delete(f"{MOCK_BASE_URL}linked-accounts/{MOCK_LINKED_ACCOUNT_ID}").mock(
        return_value=httpx.Response(404, json={"message": "Linked account not found"})
    )

    with pytest.raises(NotFoundError) as exc_info:
        client.linked_accounts.delete(MOCK_LINKED_ACCOUNT_ID)

    assert "Linked account not found" in str(exc_info.value)
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_enable_linked_account_success(client: ACI) -> None:
    mock_response = {
        "id": MOCK_LINKED_ACCOUNT_ID,
        "project_id": MOCK_PROJECT_ID,
        "app_name": MOCK_APP_NAME,
        "linked_account_owner_id": MOCK_OWNER_ID,
        "security_scheme": SecurityScheme.API_KEY,
        "enabled": True,
        "created_at": "2023-01-01T00:00:00Z",
        "updated_at": "2023-01-01T00:00:00Z",
    }

    route = respx.patch(f"{MOCK_BASE_URL}linked-accounts/{MOCK_LINKED_ACCOUNT_ID}").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    result = client.linked_accounts.enable(MOCK_LINKED_ACCOUNT_ID)

    assert result.id == UUID(MOCK_LINKED_ACCOUNT_ID)
    assert result.enabled is True
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_disable_linked_account_success(client: ACI) -> None:
    mock_response = {
        "id": MOCK_LINKED_ACCOUNT_ID,
        "project_id": MOCK_PROJECT_ID,
        "app_name": MOCK_APP_NAME,
        "linked_account_owner_id": MOCK_OWNER_ID,
        "security_scheme": SecurityScheme.API_KEY,
        "enabled": False,
        "created_at": "2023-01-01T00:00:00Z",
        "updated_at": "2023-01-01T00:00:00Z",
    }

    route = respx.patch(f"{MOCK_BASE_URL}linked-accounts/{MOCK_LINKED_ACCOUNT_ID}").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    result = client.linked_accounts.disable(MOCK_LINKED_ACCOUNT_ID)

    assert result.id == UUID(MOCK_LINKED_ACCOUNT_ID)
    assert result.enabled is False
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_server_error_retry(client: ACI) -> None:
    mock_success_response = {
        "id": MOCK_LINKED_ACCOUNT_ID,
        "project_id": MOCK_PROJECT_ID,
        "app_name": MOCK_APP_NAME,
        "linked_account_owner_id": MOCK_OWNER_ID,
        "security_scheme": SecurityScheme.API_KEY,
        "enabled": True,
        "created_at": "2023-01-01T00:00:00Z",
        "updated_at": "2023-01-01T00:00:00Z",
    }

    # Simulate two server errors followed by a successful response
    route = respx.get(f"{MOCK_BASE_URL}linked-accounts/{MOCK_LINKED_ACCOUNT_ID}").mock(
        side_effect=[
            httpx.Response(500, json={"message": "Internal server error"}),
            httpx.Response(500, json={"message": "Internal server error"}),
            httpx.Response(200, json=mock_success_response),
        ]
    )

    linked_account = client.linked_accounts.get(MOCK_LINKED_ACCOUNT_ID)

    assert route.call_count == 3, "should retry until success"
    assert linked_account.id == UUID(MOCK_LINKED_ACCOUNT_ID)


@respx.mock
def test_rate_limit_retry(client: ACI) -> None:
    mock_success_response = {
        "id": MOCK_LINKED_ACCOUNT_ID,
        "project_id": MOCK_PROJECT_ID,
        "app_name": MOCK_APP_NAME,
        "linked_account_owner_id": MOCK_OWNER_ID,
        "security_scheme": SecurityScheme.API_KEY,
        "enabled": True,
        "created_at": "2023-01-01T00:00:00Z",
        "updated_at": "2023-01-01T00:00:00Z",
    }

    # Simulate rate limit error followed by a successful response
    route = respx.get(f"{MOCK_BASE_URL}linked-accounts/{MOCK_LINKED_ACCOUNT_ID}").mock(
        side_effect=[
            httpx.Response(429, json={"message": "Rate limit exceeded"}),
            httpx.Response(200, json=mock_success_response),
        ]
    )

    linked_account = client.linked_accounts.get(MOCK_LINKED_ACCOUNT_ID)

    assert route.call_count == 2, "should retry after rate limit"
    assert linked_account.id == UUID(MOCK_LINKED_ACCOUNT_ID)



from __future__ import annotations

import logging
import os
from types import TracebackType
from typing import Any

import httpx

from aci._constants import DEFAULT_SERVER_URL
from aci._exceptions import APIKeyNotFound
from aci.meta_functions import (
    ACIExecuteFunction,
    ACIGetFunctionDefinition,
    ACISearchApps,
    ACISearchFunctions,
    ACISearchFunctionsWithIntent,
)
from aci.resource.app_configurations import AppConfigurationsResource
from aci.resource.apps import AppsResource
from aci.resource.functions import FunctionsResource
from aci.resource.linked_accounts import LinkedAccountsResource
from aci.types.functions import FunctionDefinitionFormat

logger: logging.Logger = logging.getLogger(__name__)


class ACI:
    """Client for interacting with the Aipolabs ACI API.

    This class provides methods to interact with various Aipolabs ACI endpoints,
    including searching apps and functions, getting function definitions, and
    executing functions.

    Attributes:
        api_key (str): The API key used for authentication.
        base_url (str | httpx.URL): The base URL for API requests.
        headers (dict): HTTP headers used in requests.
        client (httpx.Client): The HTTP client for making requests.
    """

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | httpx.URL | None = None,
    ) -> None:
        """Create and initialize a new Aipolabs ACI client.

        Args:
            api_key: The API key to use for authentication.
            base_url: The base URL to use for the API requests.
            If values are not provided it will try to read from the corresponding environment variables.
            If no value found for api_key, it will raise APIKeyNotFound.
            If no value found for base_url, it will use the default value.
        """
        if api_key is None:
            api_key = os.environ.get("ACI_API_KEY")
        if api_key is None:
            raise APIKeyNotFound("The API key is not found.")
        self.api_key = api_key

        if base_url is None:
            base_url = os.environ.get("AIPOLABS_ACI_SERVER_URL", DEFAULT_SERVER_URL)
        self.base_url = self._enforce_trailing_slash(httpx.URL(base_url))
        self.headers = {
            "Content-Type": "application/json",
            "x-api-key": api_key,
        }
        self.httpx_client = httpx.Client(base_url=self.base_url, headers=self.headers)

        # Initialize resource clients
        self.apps = AppsResource(self.httpx_client)
        self.functions = FunctionsResource(self.httpx_client)
        self.app_configurations = AppConfigurationsResource(self.httpx_client)
        self.linked_accounts = LinkedAccountsResource(self.httpx_client)

    def __enter__(self) -> ACI:
        self.httpx_client.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        self.httpx_client.__exit__(exc_type, exc_val, exc_tb)

    def handle_function_call(
        self,
        function_name: str,
        function_arguments: dict,
        linked_account_owner_id: str,
        allowed_apps_only: bool = False,
        format: FunctionDefinitionFormat = FunctionDefinitionFormat.OPENAI,
    ) -> Any:
        """Routes and executes function calls based on the function name.
        This can be a convenience function to handle function calls from LLM without you checking the function name.

        It supports handling built-in meta functions (ACI_SEARCH_APPS, ACI_SEARCH_FUNCTIONS,
        ACI_GET_FUNCTION_DEFINITION, ACI_EXECUTE_FUNCTION) and also handling executing third-party functions
        directly like BRAVE_SEARCH__WEB_SEARCH.

        Args:
            function_name: Name of the function to be called.
            function_arguments: Dictionary containing the input arguments for the function.
            linked_account_owner_id: To specify the end-user (account owner) on behalf of whom you want to execute functions
            You need to first link corresponding account with the same owner id in the Aipolabs ACI dashboard.
            allowed_apps_only: If true, only returns functions/apps that are allowed to be used by the agent/accessor, identified by the api key.
            format: Decides the function definition format returned by 'functions.get_definition' and 'functions.search'
        Returns:
            Any: The result (serializable) of the function execution. It varies based on the function.
        """
        logger.info(
            f"Handling function call with "
            f"name={function_name}, "
            f"params={function_arguments}, "
            f"linked_account_owner_id={linked_account_owner_id}, "
            f"allowed_apps_only={allowed_apps_only}, "
            f"format={format}"
        )
        if function_name == ACISearchApps.NAME:
            apps = self.apps.search(**function_arguments, allowed_apps_only=allowed_apps_only)

            return [app.model_dump(exclude_none=True) for app in apps]

        elif (
            function_name == ACISearchFunctions.NAME
            or function_name == ACISearchFunctionsWithIntent.NAME
        ):
            functions = self.functions.search(
                **function_arguments,
                allowed_apps_only=allowed_apps_only,
                format=format,
            )

            return functions

        elif function_name == ACIGetFunctionDefinition.NAME:
            return self.functions.get_definition(**function_arguments, format=format)

        elif function_name == ACIExecuteFunction.NAME:
            # TODO: sometimes when using the fixed_tool approach llm most time doesn't put input arguments in the
            # 'function_arguments' key as defined in ACI_EXECUTE_FUNCTION schema,
            # so we need to handle that here. It is a bit hacky, we should improve this in the future
            # TODO: consider adding post processing to auto fix all common errors in llm generated input arguments
            function_arguments = ACIExecuteFunction.wrap_function_arguments_if_not_present(
                function_arguments
            )
            result = self.functions.execute(
                **function_arguments, linked_account_owner_id=linked_account_owner_id
            )
            return result.model_dump(exclude_none=True)

        else:
            # If the function name is not a meta function, we assume it is a direct function execution of
            # an aipolabs indexed function
            # TODO: check function exist if not throw excpetion?
            result = self.functions.execute(
                function_name, function_arguments, linked_account_owner_id
            )
            return result.model_dump(exclude_none=True)

    def _enforce_trailing_slash(self, url: httpx.URL) -> httpx.URL:
        """Ensures the URL ends with a trailing slash.

        Args:
            url: The URL to process.

        Returns:
            httpx.URL: URL with a guaranteed trailing slash.
        """
        if url.raw_path.endswith(b"/"):
            return url
        return url.copy_with(raw_path=url.raw_path + b"/")



"""
This module defines the ACI_SEARCH_APPS meta function, which is used by LLM to search for
relevant applications (which include a set of functions) that can help complete a task.
"""

NAME = "ACI_SEARCH_APPS"

SCHEMA = {
    "type": "function",
    "function": {
        "name": NAME,
        "description": "This function allows you to find relevant apps (which includeds a set of functions) "
        "that can help complete your tasks or get data and information you need.",
        "parameters": {
            "type": "object",
            "properties": {
                "intent": {
                    "type": "string",
                    "description": "Use this to find relevant apps you might need. Returned results of this "
                    "function will be sorted by relevance to the intent. Examples include 'what's "
                    "the top news in the stock market today', 'i want to automate outbound "
                    "marketing emails'.",
                },
                "limit": {
                    "type": "integer",
                    "default": 100,
                    "description": "The maximum number of apps to return from the search.",
                    "minimum": 1,
                    "maximum": 1000,
                },
                "offset": {
                    "type": "integer",
                    "default": 0,
                    "minimum": 0,
                    "description": "Pagination offset.",
                },
            },
            "required": [],
            "additionalProperties": False,
        },
    },
}



import uuid
from datetime import datetime

import httpx
import pytest
import respx

from aci import ACI
from aci.types.app_configurations import AppConfiguration
from aci.types.enums import SecurityScheme

from .utils import MOCK_BASE_URL


@respx.mock
@pytest.mark.parametrize(
    "list_params",
    [
        {},  # Base case: No optional parameters provided.
        {  # All optional parameters provided.
            "app_names": ["app1", "app2"],
            "limit": 10,
            "offset": 5,
        },
    ],
)
def test_list_app_configurations_success(client: ACI, list_params: dict) -> None:
    mock_response = [
        {
            "id": str(uuid.uuid4()),
            "project_id": str(uuid.uuid4()),
            "app_name": "app1",
            "security_scheme": "oauth2",
            "security_scheme_overrides": {},
            "enabled": True,
            "all_functions_enabled": True,
            "enabled_functions": [],
            "created_at": datetime.now().isoformat(),
            "updated_at": datetime.now().isoformat(),
        }
    ]

    route = respx.get(f"{MOCK_BASE_URL}app-configurations").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    app_configs = client.app_configurations.list(**list_params)
    assert len(app_configs) == len(mock_response)
    assert app_configs[0].app_name == mock_response[0]["app_name"]
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_get_app_configuration_success(client: ACI) -> None:
    app_name = "test_app"
    mock_response = {
        "id": str(uuid.uuid4()),
        "project_id": str(uuid.uuid4()),
        "app_name": app_name,
        "security_scheme": "oauth2",
        "security_scheme_overrides": {},
        "enabled": True,
        "all_functions_enabled": True,
        "enabled_functions": [],
        "created_at": datetime.now().isoformat(),
        "updated_at": datetime.now().isoformat(),
    }

    route = respx.get(f"{MOCK_BASE_URL}app-configurations/{app_name}").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    app_config = client.app_configurations.get(app_name)
    assert isinstance(app_config, AppConfiguration)
    assert app_config.app_name == app_name
    assert route.call_count == 1, "should not retry"


# TODO: add coverage for security_scheme_overrides, all_functions_enabled, enabled_functions
@respx.mock
def test_create_app_configuration_success(client: ACI) -> None:
    app_name = "test_app"
    security_scheme = SecurityScheme.OAUTH2

    mock_response = {
        "id": str(uuid.uuid4()),
        "project_id": str(uuid.uuid4()),
        "app_name": app_name,
        "security_scheme": security_scheme,
        "security_scheme_overrides": {},
        "enabled": True,
        "all_functions_enabled": True,
        "enabled_functions": [],
        "created_at": datetime.now().isoformat(),
        "updated_at": datetime.now().isoformat(),
    }

    route = respx.post(f"{MOCK_BASE_URL}app-configurations").mock(
        return_value=httpx.Response(200, json=mock_response)
    )

    app_config = client.app_configurations.create(
        app_name,
        security_scheme,
    )
    assert isinstance(app_config, AppConfiguration)
    assert app_config.app_name == app_name
    assert app_config.security_scheme == security_scheme
    assert route.call_count == 1, "should not retry"


@respx.mock
def test_delete_app_configuration_success(client: ACI) -> None:
    app_name = "test_app"

    route = respx.delete(f"{MOCK_BASE_URL}app-configurations/{app_name}").mock(
        return_value=httpx.Response(204)
    )

    client.app_configurations.delete(app_name)
    assert route.call_count == 1, "should not retry"



import logging

from tenacity import retry

from aci.resource._base import APIResource, retry_config
from aci.types.apps import AppBasic, AppDetails, SearchAppsParams

logger: logging.Logger = logging.getLogger(__name__)


class AppsResource(APIResource):
    @retry(**retry_config)  # type: ignore
    def search(
        self,
        intent: str | None = None,
        allowed_apps_only: bool = False,
        include_functions: bool = False,
        categories: list[str] | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[AppBasic]:
        """Search for apps.

        Args:
            intent: search results will be sorted by relevance to this intent.
            allowed_apps_only: If true, only return apps that are allowed by the agent/accessor, identified by the api key.
            include_functions: If true, include functions (name and description) in the search results.
            categories: list of categories to filter apps by.
            limit: for pagination, maximum number of apps to return.
            offset: for pagination, number of apps to skip before returning results.

        Returns:
            list[AppBasic]: List of apps matching the search criteria in the order of relevance.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        validated_params = SearchAppsParams(
            intent=intent,
            allowed_apps_only=allowed_apps_only,
            include_functions=include_functions,
            categories=categories,
            limit=limit,
            offset=offset,
        ).model_dump(exclude_none=True, mode="json")

        logger.info(f"Searching apps with params: {validated_params}")
        response = self._httpx_client.get(
            "apps/search",
            params=validated_params,
        )

        data: list[dict] = self._handle_response(response)
        apps = [AppBasic.model_validate(app) for app in data]

        return apps

    @retry(**retry_config)  # type: ignore
    def get(self, app_name: str) -> AppDetails:
        """Gets detailed information about an app."""
        response = self._httpx_client.get(f"apps/{app_name}")
        data: dict = self._handle_response(response)
        app_details: AppDetails = AppDetails.model_validate(data)
        return app_details



"""
This module defines the ACI_SEARCH_FUNCTIONS_WITH_INTENT meta function, which is used by LLM to search for
relevant executable functions that can help complete a task.

This function is similar to ACI_SEARCH_FUNCTIONS, but it removes the app_names parameter for a
simplified workflow (Otherwise you'd have to call ACI_SEARCH_APPS first to get the app names).
"""

NAME = "ACI_SEARCH_FUNCTIONS_WITH_INTENT"
SCHEMA = {
    "type": "function",
    "function": {
        "name": NAME,
        "description": "This function allows you to find relevant executable functions (tools) that can help "
        "complete your tasks or get data and information you need.",
        "parameters": {
            "type": "object",
            "properties": {
                "intent": {
                    "type": "string",
                    "description": "Use this to discover relevant functions (tools) you might need. Returned "
                    "results of this function will be sorted by relevance to the intent. Examples "
                    "include 'what's the top news in the stock market today', 'i want to "
                    "automate outbound marketing emails'.",
                },
                "limit": {
                    "type": "integer",
                    "default": 100,
                    "description": "The maximum number of functions to return from the search per response.",
                    "minimum": 1,
                    "maximum": 1000,
                },
                "offset": {
                    "type": "integer",
                    "default": 0,
                    "minimum": 0,
                    "description": "Pagination offset.",
                },
            },
            "required": [],
            "additionalProperties": False,
        },
    },
}



from __future__ import annotations

import pytest
from pydantic import BaseModel, ConfigDict
from typing_extensions import TypedDict

from aci import to_json_schema
from aci.types.functions import FunctionDefinitionFormat


class RequiredDict(TypedDict):
    required_int: int
    optional_str: str | None


class RequiredModel(BaseModel):
    required_int: int
    optional_str: str | None


class AllowExtraArgs(BaseModel):
    ...
    model_config = ConfigDict(extra="allow")


EXPECTED_NAME = "my_function"
EXPECTED_DESCRIPTION = "This is a test function."
EXPECTED_PARAMETERS = {
    "properties": {
        "required_int": {
            "description": "This is required_int.",
            "title": "Required Int",
            "type": "integer",
        },
        "required_dict": {
            "description": "This is required_dict.",
            "properties": {
                "required_int": {"title": "Required Int", "type": "integer"},
                "optional_str": {
                    "anyOf": [{"type": "string"}, {"type": "null"}],
                    "title": "Optional Str",
                },
            },
            "required": ["required_int", "optional_str"],
            "title": "RequiredDict",
            "type": "object",
            "additionalProperties": False,
        },
        "required_model": {
            "description": "This is required_model.",
            "properties": {
                "required_int": {"title": "Required Int", "type": "integer"},
                "optional_str": {
                    "anyOf": [{"type": "string"}, {"type": "null"}],
                    "title": "Optional Str",
                },
            },
            "required": ["required_int", "optional_str"],
            "title": "RequiredModel",
            "type": "object",
            "additionalProperties": False,
        },
        "allow_extra_args": {
            "description": "This is allow_extra_args.",
            "additionalProperties": True,
            "properties": {},
            "title": "AllowExtraArgs",
            "type": "object",
        },
        "optional_str_with_default": {
            "default": "default string",
            "description": "This is optional_str_with_default.",
            "title": "Optional Str With Default",
            "type": "string",
        },
    },
    "required": ["required_int", "required_dict", "required_model", "allow_extra_args"],
    "title": "my_function_args",
    "type": "object",
    "additionalProperties": False,
}

EXPECTED_JSON_SCHEMA_OPENAI_RESPONSES = {
    "type": "function",
    "name": EXPECTED_NAME,
    "description": EXPECTED_DESCRIPTION,
    "parameters": EXPECTED_PARAMETERS,
}

EXPECTED_JSON_SCHEMA_OPENAI = {
    "type": "function",
    "function": {
        "name": EXPECTED_NAME,
        "description": EXPECTED_DESCRIPTION,
        "parameters": EXPECTED_PARAMETERS,
    },
}

EXPECTED_JSON_SCHEMA_ANTHROPIC = {
    "name": EXPECTED_NAME,
    "description": EXPECTED_DESCRIPTION,
    "input_schema": EXPECTED_PARAMETERS,
}


@pytest.mark.parametrize(
    ["format", "expected_schema"],
    [
        (FunctionDefinitionFormat.OPENAI_RESPONSES, EXPECTED_JSON_SCHEMA_OPENAI_RESPONSES),
        (FunctionDefinitionFormat.OPENAI, EXPECTED_JSON_SCHEMA_OPENAI),
        (FunctionDefinitionFormat.ANTHROPIC, EXPECTED_JSON_SCHEMA_ANTHROPIC),
    ],
)
def test_schema_conversion(format: FunctionDefinitionFormat, expected_schema: dict) -> None:
    # dummy function to test the schema conversion
    def my_function(
        required_int: int,
        required_dict: RequiredDict,
        required_model: RequiredModel,
        allow_extra_args: AllowExtraArgs,
        optional_str_with_default: str = "default string",
    ) -> None:
        """This is a test function.

        Args:
            required_int: This is required_int.
            required_dict: This is required_dict.
            required_model: This is required_model.
            allow_extra_args: This is allow_extra_args.
            optional_str_with_default: This is optional_str_with_default.
        """
        pass

    # test that the schema is generated correctly
    schema = to_json_schema(my_function, format=format)
    assert schema == expected_schema



import os

import httpx
import pytest

from aci import ACI
from aci._constants import DEFAULT_SERVER_URL
from aci._exceptions import APIKeyNotFound

from .utils import MOCK_API_KEY, MOCK_BASE_URL


def test_client_initialization() -> None:
    client = ACI(api_key=MOCK_API_KEY, base_url=MOCK_BASE_URL)
    assert client.api_key == MOCK_API_KEY
    assert client.base_url == httpx.URL(MOCK_BASE_URL)
    assert client.headers["x-api-key"] == MOCK_API_KEY
    assert hasattr(client, "apps")
    assert hasattr(client, "functions")


def test_client_initialization_without_api_key() -> None:
    with pytest.raises(APIKeyNotFound):
        # unset ACI_API_KEY environment variable
        os.environ.pop("ACI_API_KEY", None)
        ACI(api_key=None, base_url=MOCK_BASE_URL)


def test_client_initialization_without_base_url() -> None:
    client = ACI(api_key=MOCK_API_KEY, base_url=None)
    assert client.base_url == httpx.URL(DEFAULT_SERVER_URL)



from datetime import datetime
from uuid import UUID

from pydantic import BaseModel, ConfigDict

from aci.types.enums import SecurityScheme


class LinkedAccountCreateBase(BaseModel):
    """Base model for creating a linked account."""

    app_name: str
    linked_account_owner_id: str


class LinkedAccountOAuth2Create(LinkedAccountCreateBase):
    """Model for creating an OAuth2 linked account."""

    after_oauth2_link_redirect_url: str | None = None


class LinkedAccountAPIKeyCreate(LinkedAccountCreateBase):
    """Model for creating an API key linked account."""

    api_key: str


class LinkedAccountNoAuthCreate(LinkedAccountCreateBase):
    """Model for creating a no-auth linked account."""

    pass


class LinkedAccountUpdate(BaseModel):
    """Model for updating a linked account."""

    enabled: bool | None = None


class LinkedAccount(BaseModel):
    """Public representation of a linked account."""

    id: UUID
    project_id: UUID
    app_name: str
    linked_account_owner_id: str
    security_scheme: SecurityScheme
    enabled: bool
    created_at: datetime
    updated_at: datetime

    model_config = ConfigDict(from_attributes=True, extra="allow")


class LinkedAccountsList(BaseModel):
    """Parameters for listing linked accounts."""

    app_name: str | None = None
    linked_account_owner_id: str | None = None



import logging

import httpx
from tenacity import retry

from aci.resource._base import APIResource, retry_config
from aci.types.functions import (
    FunctionDefinitionFormat,
    FunctionExecutionParams,
    FunctionExecutionResult,
    GetFunctionDefinitionParams,
    SearchFunctionsParams,
)

logger: logging.Logger = logging.getLogger(__name__)


class FunctionsResource(APIResource):
    def __init__(self, httpx_client: httpx.Client) -> None:
        super().__init__(httpx_client)

    @retry(**retry_config)  # type: ignore
    def search(
        self,
        app_names: list[str] | None = None,
        intent: str | None = None,
        allowed_apps_only: bool = False,
        format: FunctionDefinitionFormat = FunctionDefinitionFormat.OPENAI,
        limit: int | None = None,
        offset: int | None = None,
    ) -> list[dict]:
        """Searches for functions.
        # TODO: return specific pydantic model for returned functions based on FunctionDefinitionFormat

        Args:
            app_names: List of app names to filter functions by.
            intent: search results will be sorted by relevance to this intent.
            allowed_apps_only: If true, only returns functions of apps that are allowed by the
                agent/accessor, identified by the api key.
            limit: for pagination, maximum number of functions to return.
            offset: for pagination, number of functions to skip before returning results.

        Returns:
            list[dict]: List of functions matching the search criteria in the order of relevance.
            The format of the functions is determined by the FunctionDefinitionFormat.
        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        validated_params = SearchFunctionsParams(
            app_names=app_names,
            intent=intent,
            allowed_apps_only=allowed_apps_only,
            format=format,
            limit=limit,
            offset=offset,
        ).model_dump(exclude_none=True, mode="json")

        logger.info(f"Searching functions with params: {validated_params}")
        response = self._httpx_client.get(
            "functions/search",
            params=validated_params,
        )

        data: list[dict] = self._handle_response(response)

        return data

    @retry(**retry_config)  # type: ignore
    def get_definition(
        self, function_name: str, format: FunctionDefinitionFormat = FunctionDefinitionFormat.OPENAI
    ) -> dict:
        """Retrieves the definition of a specific function.
        # TODO: return specific pydantic model for returned functions based on FunctionDefinitionFormat

        Args:
            function_name: Name of the function to retrieve.
            format: Decide the function definition format.

        Returns:
            # TODO: specific pydantic model for returned function definition based on FunctionDefinitionFormat
            dict: JSON schema that defines the function, varies based on the FunctionDefinitionFormat.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        validated_params = GetFunctionDefinitionParams(function_name=function_name, format=format)

        logger.info(
            f"Getting function definition of {validated_params.function_name}, "
            f"format: {validated_params.format}"
        )
        response = self._httpx_client.get(
            f"functions/{validated_params.function_name}/definition",
            params={"format": validated_params.format.value},
        )

        function_definition: dict = self._handle_response(response)

        return function_definition

    @retry(**retry_config)  # type: ignore
    def execute(
        self, function_name: str, function_arguments: dict, linked_account_owner_id: str
    ) -> FunctionExecutionResult:
        """Executes a Aipolabs ACI indexed function with the provided arguments.

        Args:
            function_name: Name of the function to execute.
            function_arguments: Dictionary containing the input arguments for the function.
            linked_account_owner_id: to specify with credentials of which linked account the
                function should be executed.
        Returns:
            FunctionExecutionResult: containing the function execution results.

        Raises:
            Various exceptions defined in _handle_response for different HTTP status codes.
        """
        validated_params = FunctionExecutionParams(
            function_name=function_name,
            function_arguments=function_arguments,
            linked_account_owner_id=linked_account_owner_id,
        )

        logger.info(f"Executing function with: {validated_params.model_dump()}")
        request_body = {
            "function_input": validated_params.function_arguments,
            "linked_account_owner_id": validated_params.linked_account_owner_id,
        }
        response = self._httpx_client.post(
            f"functions/{validated_params.function_name}/execute",
            json=request_body,
        )

        function_execution_result: FunctionExecutionResult = FunctionExecutionResult.model_validate(
            self._handle_response(response)
        )

        return function_execution_result



import os
import uuid

import pytest

from aci import ACI
from aci.types.enums import SecurityScheme
from aci.types.linked_accounts import LinkedAccount

ACI_API_KEY = os.getenv("ACI_API_KEY")

"""
The codease doesn't use load_dotenv() so you need to set the environment variables manually.
You can use something like dotenv to load the environment variables from a .env file for a single run.
e.g.

```
$ npm install -g dotenv-cli
$ dotenv run pytest tests/it/test_linked_accounts.py -s
```
"""

# Test user ID to use across tests
TEST_OWNER_ID = f"test-user-{uuid.uuid4()}"


@pytest.mark.skipif(ACI_API_KEY is None, reason="ACI_API_KEY is not set")
def test_linked_accounts() -> None:
    """
    Test the linked accounts API.
    """
    aci = ACI(api_key=ACI_API_KEY)

    # step 0: make sure no app configurations exist and set up test app configurations
    assert len(aci.app_configurations.list()) == 0

    # step 1: set up app configurations with different security schemes
    aci.app_configurations.create(
        app_name="BRAVE_SEARCH",
        security_scheme=SecurityScheme.API_KEY,
    )
    aci.app_configurations.create(
        app_name="AIPOLABS_SECRETS_MANAGER",
        security_scheme=SecurityScheme.NO_AUTH,
    )

    # step 2: link accounts with different security schemes
    # Link a NO_AUTH account
    no_auth_account = aci.linked_accounts.link(
        app_name="AIPOLABS_SECRETS_MANAGER",
        linked_account_owner_id=TEST_OWNER_ID,
        security_scheme=SecurityScheme.NO_AUTH,
    )
    no_auth_account = LinkedAccount.model_validate(no_auth_account)
    assert no_auth_account.app_name == "AIPOLABS_SECRETS_MANAGER"
    assert no_auth_account.linked_account_owner_id == TEST_OWNER_ID
    assert no_auth_account.security_scheme == SecurityScheme.NO_AUTH
    assert no_auth_account.enabled is True

    # Link an API_KEY account
    # Use a placeholder API key for testing
    test_api_key = "test-api-key-" + str(uuid.uuid4())
    api_key_account = aci.linked_accounts.link(
        app_name="BRAVE_SEARCH",
        linked_account_owner_id=TEST_OWNER_ID,
        security_scheme=SecurityScheme.API_KEY,
        api_key=test_api_key,
    )
    api_key_account = LinkedAccount.model_validate(api_key_account)
    assert api_key_account.app_name == "BRAVE_SEARCH"
    assert api_key_account.linked_account_owner_id == TEST_OWNER_ID
    assert api_key_account.security_scheme == SecurityScheme.API_KEY
    assert api_key_account.enabled is True

    # Link with OAuth2 would return a URL rather than an account object
    # We're not testing this in integration tests as it requires browser interaction

    # step 3: list all linked accounts for our test user
    accounts = aci.linked_accounts.list(linked_account_owner_id=TEST_OWNER_ID)
    assert len(accounts) == 2
    assert {account.app_name for account in accounts} == {
        "BRAVE_SEARCH",
        "AIPOLABS_SECRETS_MANAGER",
    }

    # step 4: get a specific linked account
    retrieved_account = aci.linked_accounts.get(api_key_account.id)
    assert retrieved_account.id == api_key_account.id
    assert retrieved_account.app_name == "BRAVE_SEARCH"
    assert retrieved_account.linked_account_owner_id == TEST_OWNER_ID

    # step 5: list linked accounts filtered by app_name
    brave_accounts = aci.linked_accounts.list(
        app_name="BRAVE_SEARCH", linked_account_owner_id=TEST_OWNER_ID
    )
    assert len(brave_accounts) == 1
    assert brave_accounts[0].app_name == "BRAVE_SEARCH"

    # step 6: disable a linked account
    disabled_account = aci.linked_accounts.disable(api_key_account.id)
    assert disabled_account.enabled is False

    # Verify the account is disabled
    retrieved_account = aci.linked_accounts.get(api_key_account.id)
    assert retrieved_account.enabled is False

    # step 7: enable a linked account
    enabled_account = aci.linked_accounts.enable(api_key_account.id)
    assert enabled_account.enabled is True

    # Verify the account is enabled
    retrieved_account = aci.linked_accounts.get(api_key_account.id)
    assert retrieved_account.enabled is True

    # step 8: delete all linked accounts
    accounts = aci.linked_accounts.list(linked_account_owner_id=TEST_OWNER_ID)
    for account in accounts:
        aci.linked_accounts.delete(account.id)

    # Verify accounts are deleted
    accounts = aci.linked_accounts.list(linked_account_owner_id=TEST_OWNER_ID)
    assert len(accounts) == 0

    # step 9: clean up app configurations
    for app_name in ["BRAVE_SEARCH", "AIPOLABS_SECRETS_MANAGER"]:
        aci.app_configurations.delete(app_name=app_name)

    # step 10: verify app configurations are deleted
    assert len(aci.app_configurations.list()) == 0


@pytest.mark.skipif(ACI_API_KEY is None, reason="ACI_API_KEY is not set")
def test_oauth2_account_linking() -> None:
    """
    Test the linked accounts API with OAuth2.
    """
    aci = ACI(api_key=ACI_API_KEY)

    # step 0: make sure no app configurations exist and set up test app configurations
    assert len(aci.app_configurations.list()) == 0

    # step 1: create a GMAIL app configuration
    aci.app_configurations.create(
        app_name="GMAIL",
        security_scheme=SecurityScheme.OAUTH2,
    )

    # step 2: link a GMAIL account
    oauth2_account = aci.linked_accounts.link(
        app_name="GMAIL",
        linked_account_owner_id=TEST_OWNER_ID,
        security_scheme=SecurityScheme.OAUTH2,
    )
    assert isinstance(oauth2_account, str)

    # step 3: delete the GMAIL app configuration
    aci.app_configurations.delete(app_name="GMAIL")

    # step 4: verify the GMAIL app configuration is deleted
    assert len(aci.app_configurations.list()) == 0



from __future__ import annotations

from typing import Any, Literal

from typing_extensions import TypeGuard, override

"""
This file is a modified version of the `strict_schema.py` file from the `openai` library.
https://github.com/openai/openai-agents-python/blob/064e25b01b5c82c08aea66ff898ff27adbb013d8/src/agents/strict_schema.py
It is used to ensure that the JSON schema is compatible with the OpenAI API. (not to enforce `strict` schema)
"""


# Sentinel class used until PEP 0661 is accepted
class NotGiven:
    """
    A sentinel singleton class used to distinguish omitted keyword arguments
    from those passed in with the value None (which may have different behavior).

    For example:

    ```py
    def get(timeout: Union[int, NotGiven, None] = NotGiven()) -> Response: ...


    get(timeout=1)  # 1s timeout
    get(timeout=None)  # No timeout
    get()  # Default timeout behavior, which may not be statically known at the method definition.
    ```
    """

    def __bool__(self) -> Literal[False]:
        return False

    @override
    def __repr__(self) -> str:
        return "NOT_GIVEN"


NOT_GIVEN = NotGiven()

_EMPTY_SCHEMA = {
    "additionalProperties": False,
    "type": "object",
    "properties": {},
    "required": [],
}


def ensure_llm_compatible_json_schema(
    schema: dict[str, Any],
) -> dict[str, Any]:
    """Mutates the given JSON schema to ensure it conforms to the `strict` standard
    that the OpenAI API expects.
    """
    if schema == {}:
        return _EMPTY_SCHEMA
    return _ensure_llm_compatible_json_schema(schema, path=(), root=schema)


# Adapted from https://github.com/openai/openai-python/blob/main/src/openai/lib/_pydantic.py
def _ensure_llm_compatible_json_schema(
    json_schema: object,
    *,
    path: tuple[str, ...],
    root: dict[str, object],
) -> dict[str, Any]:
    if not is_dict(json_schema):
        raise TypeError(f"Expected {json_schema} to be a dictionary; path={path}")

    defs = json_schema.get("$defs")
    if is_dict(defs):
        for def_name, def_schema in defs.items():
            _ensure_llm_compatible_json_schema(
                def_schema, path=(*path, "$defs", def_name), root=root
            )

    definitions = json_schema.get("definitions")
    if is_dict(definitions):
        for definition_name, definition_schema in definitions.items():
            _ensure_llm_compatible_json_schema(
                definition_schema, path=(*path, "definitions", definition_name), root=root
            )

    typ = json_schema.get("type")
    if typ == "object" and "additionalProperties" not in json_schema:
        json_schema["additionalProperties"] = False

    # object types
    # { 'type': 'object', 'properties': { 'a':  {...} } }
    properties = json_schema.get("properties")
    if is_dict(properties):
        # NOTE: we don't want this (All fields in properties must be marked as required)
        # because we are using the `strict_mode=False` only for openai schema
        # json_schema["required"] = list(properties.keys())
        json_schema["properties"] = {
            key: _ensure_llm_compatible_json_schema(
                prop_schema, path=(*path, "properties", key), root=root
            )
            for key, prop_schema in properties.items()
        }

    # arrays
    # { 'type': 'array', 'items': {...} }
    items = json_schema.get("items")
    if is_dict(items):
        json_schema["items"] = _ensure_llm_compatible_json_schema(
            items, path=(*path, "items"), root=root
        )

    # unions
    any_of = json_schema.get("anyOf")
    if is_list(any_of):
        json_schema["anyOf"] = [
            _ensure_llm_compatible_json_schema(variant, path=(*path, "anyOf", str(i)), root=root)
            for i, variant in enumerate(any_of)
        ]

    # intersections
    all_of = json_schema.get("allOf")
    if is_list(all_of):
        if len(all_of) == 1:
            json_schema.update(
                _ensure_llm_compatible_json_schema(all_of[0], path=(*path, "allOf", "0"), root=root)
            )
            json_schema.pop("allOf")
        else:
            json_schema["allOf"] = [
                _ensure_llm_compatible_json_schema(entry, path=(*path, "allOf", str(i)), root=root)
                for i, entry in enumerate(all_of)
            ]

    # strip `None` defaults as there's no meaningful distinction here
    # the schema will still be `nullable` and the model will default
    # to using `None` anyway
    if json_schema.get("default", NOT_GIVEN) is None:
        json_schema.pop("default")

    # we can't use `$ref`s if there are also other properties defined, e.g.
    # `{"$ref": "...", "description": "my description"}`
    #
    # so we unravel the ref
    # `{"type": "string", "description": "my description"}`
    ref = json_schema.get("$ref")
    # NOTE: opinionated change: we expand refs regardless of the number of keys. previously, we only expanded refs
    # if there was only one key in the json schema. (has_more_than_n_keys(json_schema, 1))
    if ref:
        assert isinstance(ref, str), f"Received non-string $ref - {ref}"

        resolved = resolve_ref(root=root, ref=ref)
        if not is_dict(resolved):
            raise ValueError(
                f"Expected `$ref: {ref}` to resolved to a dictionary but got {resolved}"
            )

        # properties from the json schema take priority over the ones on the `$ref`
        json_schema.update({**resolved, **json_schema})
        json_schema.pop("$ref")
        # Since the schema expanded from `$ref` might not have `additionalProperties: false` applied
        # we call `_ensure_llm_compatible_json_schema` again to fix the inlined schema and ensure it's valid
        return _ensure_llm_compatible_json_schema(json_schema, path=path, root=root)

    return json_schema


def resolve_ref(*, root: dict[str, object], ref: str) -> object:
    if not ref.startswith("#/"):
        raise ValueError(f"Unexpected $ref format {ref!r}; Does not start with #/")

    path = ref[2:].split("/")
    resolved = root
    for key in path:
        value = resolved[key]
        assert is_dict(value), (
            f"encountered non-dictionary entry while resolving {ref} - {resolved}"
        )
        resolved = value

    return resolved


def is_dict(obj: object) -> TypeGuard[dict[str, object]]:
    # just pretend that we know there are only `str` keys
    # as that check is not worth the performance cost
    return isinstance(obj, dict)


def is_list(obj: object) -> TypeGuard[list[object]]:
    return isinstance(obj, list)


def has_more_than_n_keys(obj: dict[str, object], n: int) -> bool:
    i = 0
    for _ in obj.keys():
        i += 1
        if i > n:
            return True
    return False



from datetime import datetime
from uuid import UUID

from pydantic import BaseModel, ConfigDict

from aci.types.enums import SecurityScheme


class AppConfiguration(BaseModel):
    """Public representation of an app configuration."""

    id: UUID
    project_id: UUID
    app_name: str
    security_scheme: SecurityScheme
    enabled: bool
    all_functions_enabled: bool
    enabled_functions: list[str]
    created_at: datetime
    updated_at: datetime

    # allow extra fields for backward compatibility
    model_config = ConfigDict(extra="allow")


class AppConfigurationCreate(BaseModel):
    """Create a new app configuration.

    "all_functions_enabled=True" → ignore enabled_functions.
    "all_functions_enabled=False" AND non-empty enabled_functions → selectively enable that list.
    "all_functions_enabled=False" AND empty enabled_functions → all functions disabled.
    """

    app_name: str
    security_scheme: SecurityScheme
    security_scheme_overrides: dict | None = None
    all_functions_enabled: bool = True
    enabled_functions: list[str] | None = None


# TODO: add limit and offset boundaries
class AppConfigurationsList(BaseModel):
    app_names: list[str] | None = None
    limit: int | None = None
    offset: int | None = None



import logging
from typing import Any

import httpx
from tenacity import (
    after_log,
    before_log,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from aci._constants import (
    DEFAULT_MAX_RETRIES,
    DEFAULT_RETRY_MAX_WAIT,
    DEFAULT_RETRY_MIN_WAIT,
    DEFAULT_RETRY_MULTIPLIER,
)
from aci._exceptions import (
    AuthenticationError,
    NotFoundError,
    PermissionError,
    RateLimitError,
    ServerError,
    UnknownError,
    ValidationError,
)

logger: logging.Logger = logging.getLogger(__name__)


class APIResource:
    _httpx_client: httpx.Client

    def __init__(self, httpx_client: httpx.Client) -> None:
        self._httpx_client = httpx_client

    def _handle_response(self, response: httpx.Response) -> Any:
        """Processes API responses and handles errors.

        Args:
            response: The HTTP response from the API.

        Returns:
            Any: Parsed JSON response for successful requests.

        Raises:
            AuthenticationError: For 401 status codes.
            PermissionError: For 403 status codes.
            NotFoundError: For 404 status codes.
            ValidationError: For 400 status codes.
            RateLimitError: For 429 status codes.
            ServerError: For 5xx status codes.
            UnknownError: For unexpected status codes.
        """

        try:
            response.raise_for_status()
            return self._get_response_data(response)

        except httpx.HTTPStatusError as e:
            error_message = self._get_error_message(response, e)

            # TODO: cross-check with backend
            if response.status_code == 401:
                raise AuthenticationError(error_message) from e
            elif response.status_code == 403:
                raise PermissionError(error_message) from e
            elif response.status_code == 404:
                raise NotFoundError(error_message) from e
            elif response.status_code == 400:
                raise ValidationError(error_message) from e
            elif response.status_code == 429:
                raise RateLimitError(error_message) from e
            elif 500 <= response.status_code < 600:
                raise ServerError(error_message) from e
            else:
                raise UnknownError(error_message) from e

    def _get_response_data(self, response: httpx.Response) -> Any:
        """Get the response data from the response.
        If the response is json, return the json data, otherwise fallback to the text.
        TODO: handle non-json response?
        """
        try:
            response_data = response.json() if response.content else {}
        except Exception as e:
            logger.warning(f"error parsing json response: {e!s}")
            response_data = response.text

        return response_data

    def _get_error_message(self, response: httpx.Response, error: httpx.HTTPStatusError) -> str:
        """Get the error message from the response or fallback to the error message from the HTTPStatusError.
        Usually the response json contains more details about the error.
        """
        try:
            return str(response.json())
        except Exception:
            return str(error)


# Shared retry config for all requests to the Aipolabs ACI API
retry_config = {
    "stop": stop_after_attempt(DEFAULT_MAX_RETRIES),
    "wait": wait_exponential(
        multiplier=DEFAULT_RETRY_MULTIPLIER,
        min=DEFAULT_RETRY_MIN_WAIT,
        max=DEFAULT_RETRY_MAX_WAIT,
    ),
    "retry": retry_if_exception_type(
        (
            ServerError,
            RateLimitError,
            UnknownError,
            httpx.TimeoutException,
            httpx.NetworkError,
        )
    ),
    "before": before_log(logger, logging.DEBUG),
    "after": after_log(logger, logging.DEBUG),
    "reraise": True,
}
